/*=============================================================================
author        :Walter Schreppers
filename      :matrices.h
description   :functions used by matlab to C++ compiled scripts
bugreport(log):/
=============================================================================*/

#include "numtypes.h"
#include "matrix.h"
#include <iostream>
#include <string>
#include <map>
#include <boost/shared_ptr.hpp>

#include "tmatrix.h"
#include "mpmatrix.h"
#include "ratmatrix.h"
#include "cmpmatrix.h"
#include "strmatrix.h"
#include "dmatrix.h"
#include "bmatrix.h"

using namespace std;

typedef boost::shared_ptr<Matrix> MatrixPtr;
typedef map< string , MatrixPtr > symtable;
numTypes currentMode = tMpMatrix;
string verbose;


numTypes getMode( MatrixPtr& arg ){
  if( dynamic_cast< StrMatrix* >( arg.get() ) != 0 ){
    return tString;
  }
  else if( dynamic_cast< CMpMatrix* >( arg.get() ) != 0 ){
    return tCMpMatrix;
  }
  else if( dynamic_cast< MpMatrix* >( arg.get() ) != 0 ){
    return tMpMatrix;
  }
  else if( dynamic_cast< RatMatrix* >( arg.get() ) != 0 ){
    return tRatMatrix;
  }
  else if( dynamic_cast< DMatrix* >( arg.get() ) != 0 ){
    return tDMatrix;
  }
  else if( dynamic_cast< BMatrix* >( arg.get() ) != 0 ){
    return tBMatrix;
  }
  else{
    return tUnknown;
  }
}

MatrixPtr deepCopy( MatrixPtr arg ){ 
  MatrixPtr retArg; 
  numTypes argMode = getMode( arg );
  switch( argMode ){
    case tString     : retArg.reset( new StrMatrix( *arg ) ); break;
    case tMpMatrix   : retArg.reset( new MpMatrix( *arg )  ); break;
    case tCMpMatrix  : retArg.reset( new CMpMatrix( *arg ) ); break;
    case tRatMatrix  : retArg.reset( new RatMatrix( *arg ) ); break;
    case tDMatrix    : retArg.reset( new DMatrix( *arg )   ); break;
    case tBMatrix    : retArg.reset( new BMatrix( *arg )   ); break;
    case tVoid       : throw InvalidOperation("Void type not supported yet due to major refactoring" );      break;
    case tUnknown    : throw InvalidOperation("Error deepCopy from Unknown type");   break;    
    default          : throw InvalidOperation("Error deepCopy from INVALID TYPE" );  break;
  }
  
  return retArg;
}



MatrixPtr getVariable( symtable& variables, const string& name ){
  symtable::iterator variable = variables.find( name );
  
  if( variable == variables.end() ){
    throw InvalidOperation( "??? Undefined function or variable '" + name + "'" );
    MatrixPtr dummy;
    return dummy;
  }
  
  return variable->second; //shallow copy
}


//This is used to create new matrix instances using a string (for constants...)
MatrixPtr newValue( const string& value ){
  MatrixPtr arg;
  switch( currentMode ){
    case tString     : arg.reset( new StrMatrix( value) ); break;
    case tMpMatrix   : arg.reset( new MpMatrix(  value) ); break;
    case tCMpMatrix  : arg.reset( new CMpMatrix( value) ); break;
    case tRatMatrix  : arg.reset( new RatMatrix( value) ); break;
    case tDMatrix    : arg.reset( new DMatrix(   value) ); break;
    case tBMatrix    : arg.reset( new BMatrix(   value) ); break;
    case tVoid       : throw InvalidOperation("Void type not supported yet due to major refactoring" );      break;
    case tUnknown    : throw InvalidOperation("Error assigning to Unknown type");   break;    
    default          : throw InvalidOperation("Error assigning to INVALID TYPE" );  break;
  }
  return arg;
}

MatrixPtr newStrValue( const string& value ){
  MatrixPtr arg;
  arg.reset( new StrMatrix( value ) );
  return arg;
}

MatrixPtr newImaginary(){
  if( currentMode != tCMpMatrix ){
    cout<<"Warning: Setting mode to complex"<<endl;
    currentMode = tCMpMatrix;
  }

  return newValue( "(;1)" );
}



// TODO the input is not general like matlab,
// we should read it in as a string and then parse it
// like an expression. This will best be done by using the eval internal
// function when this is implemented...
MatrixPtr inputValue( MatrixPtr info ){
  cout << *info ;
  MatrixPtr arg = newValue ("");
  cin >> *arg;
  return arg;
}




MatrixPtr setOutputMode( int outmode ){
 
  switch( outmode ){
    case 0  : ArithmosIO::setIOMode( 0x00000000 ); break;  //reset flags
    case 1  : ArithmosIO::setIOMode( ArithmosIO::getIOMode() | ARITHMOS_IO_MPIEEE_DECIMAL );    break;
    case 2  : ArithmosIO::setIOMode( ArithmosIO::getIOMode() | ARITHMOS_IO_MPIEEE_BINARY );     break;
    case 3  : ArithmosIO::setIOMode( ArithmosIO::getIOMode() | ARITHMOS_IO_MPIEEE_BINREP );     break;
    case 4  : ArithmosIO::setIOMode( ArithmosIO::getIOMode() | ARITHMOS_IO_MPIEEE_HEXREP );     break;
    case 5  : ArithmosIO::setIOMode( ArithmosIO::getIOMode() | ARITHMOS_IO_MPIEEE_RATIONAL );   break;
    case 6  : ArithmosIO::setIOMode( ArithmosIO::getIOMode() | ARITHMOS_IO_MPIEEE_FLAGS);       break;
    case 7  : ArithmosIO::setIOMode( ArithmosIO::getIOMode() | ARITHMOS_IO_MPIEEE_PARAM );      break;
    case 8  : ArithmosIO::setIOMode( ArithmosIO::getIOMode() | ARITHMOS_IO_RATIONAL_DECIMAL );  break;
    case 9  : ArithmosIO::setIOMode( ArithmosIO::getIOMode() | ARITHMOS_IO_RATIONAL_BINREP );   break;
    case 10 : ArithmosIO::setIOMode( ArithmosIO::getIOMode() | ARITHMOS_IO_RATIONAL_RATIONAL ); break;
    case 11 : ArithmosIO::setIOMode( ArithmosIO::getIOMode() | ARITHMOS_IO_RATIONAL_FLAGS );    break;
    case 12 : cerr<< "Warning can't set Rational Parameter outputmode, not defined in ArithmosIO." << endl; break;
    case 13 : ArithmosIO::setIOMode( ArithmosIO::getIOMode() | ARITHMOS_IO_VERBOSE );           break;
    default : ostringstream out;
              out << outmode;  
              //throw InvalidOperation("Invalid value for outputmode, must be between 0 and 13. Value given: "+out.str()+"." );
              return newValue( "0" );
              break;
  }
  
  return newValue( "1" );
}


MatrixPtr setMode( const string& type ){
  if( type == "mpieee" ){
    currentMode = tMpMatrix;
    ArithmosIO::setIOMode( ARITHMOS_IO_MPIEEE_DECIMAL );
  }
  else if( type == "rational" ){
    currentMode = tRatMatrix;
    ArithmosIO::setIOMode( ARITHMOS_IO_RATIONAL_RATIONAL );
  }
  else if( type == "complex" ){
    currentMode = tCMpMatrix;
    ArithmosIO::setIOMode( ARITHMOS_IO_CMPIEEE_DECIMAL );
  }
  else if( type == "double" ){
    currentMode = tDMatrix;
  }
  else{
    currentMode = tUnknown;
    cerr << "Invalid mode given" << endl;
    return newValue("0");
  }
  
  return newValue("1");
}

MatrixPtr setPrecision( MatrixPtr arg ){
    if( (currentMode == tMpMatrix) || (currentMode == tCMpMatrix) ){
  
    int iPrec = (int) (*arg); //the cast to int operator has been defined in Matrix!
    
    if( iPrec > 1 ){
      MpIeee::fpEnv.setPrecision( iPrec );
    }
    else{
      throw InvalidOperation("The precision must be > 1!" );
      return newValue("0");
    }
  }
  else{
    throw InvalidOperation("PRECISION function only applies to modes 'mpieee' and 'complex'" );
    return newValue("0");
  }
  
  return newValue("1");
}

MatrixPtr setRadix( MatrixPtr arg ){
  if( (currentMode == tMpMatrix) || (currentMode == tCMpMatrix) ){
  
    int i = (int) (*arg);
  
    if( i > 1 ){
      MpIeee::fpEnv.setRadix( i );
    }
    else{
      throw InvalidOperation( "The radix must be > 1!" );
    }
  }
  else{
    throw InvalidOperation("RADIX function only applies to modes 'mpieee' and 'complex'" );
  }
  return newValue("1");
}

MatrixPtr setExponent( MatrixPtr lMat, MatrixPtr uMat ){
  if( (currentMode == tMpMatrix) || (currentMode == tCMpMatrix) ){
    int L = (int) *lMat;
    int U = (int) *uMat;
    
    if( (L<0) && (U>0) ){
      MpIeee::fpEnv.setExponentRange( L, U );
    }
    else{
      ostringstream out;
      out << " L = " << L << ", U = " << U;
      throw InvalidOperation("The arguments of the EXPONENT function are wrong, L must be < 0 and U >0 " + out.str() );
    }
  }
  else{
    throw InvalidOperation("EXPONENT function only applies to modes 'mpieee' and 'complex'" );
  }
  
  return newValue("1");
}

MatrixPtr setRounding( const string& sRound ){
  if( (currentMode == tMpMatrix) || (currentMode == tCMpMatrix) ){
    if( sRound == "nearest" ){
      MpIeee::fpEnv.setRound( FP_RN );
    }
    else if ( sRound == "up"   ){
      MpIeee::fpEnv.setRound( FP_RP );
    }
    else if ( sRound == "down" ){
      MpIeee::fpEnv.setRound( FP_RM );
    }
    else if ( sRound == "zero" ){
      MpIeee::fpEnv.setRound( FP_RZ );
    }
    else{
      throw InvalidOperation("ROUNDING function was called with invalid rounding mode" );
      return newValue("1");
    }
  }
  else{
    throw InvalidOperation("ROUNDING function only applies to modes 'mpieee' and 'complex'" );
    return newValue("1");
  }
  
  return newValue("1");
}




/* ============= not needed yet... ===================

void setBoolValue( MatrixPtr& arg, bool bval ){
  arg.reset( new BMatrix( bval ) );
}

//allocate a new empty matrix
MatrixPtr newValue(){
  MatrixPtr arg;
  switch( currentMode ){
    case tString     : arg.reset( new StrMatrix() ); break;
    case tMpMatrix   : arg.reset( new MpMatrix()  ); break;
    case tCMpMatrix  : arg.reset( new CMpMatrix() ); break;
    case tRatMatrix  : arg.reset( new RatMatrix() ); break;
    case tDMatrix    : arg.reset( new DMatrix()   ); break;
    case tBMatrix    : arg.reset( new BMatrix()   ); break;
    case tVoid       : throw InvalidOperation("Void type not supported yet due to major refactoring" );      break;
    case tUnknown    : throw InvalidOperation("Error assigning to Unknown type");   break;    
    default          : throw InvalidOperation("Error assigning to INVALID TYPE" );  break;
  }
  return arg;
}

//allocate empty matrix with a specific type
void newValue( MatrixPtr& arg, numTypes mode ){
  switch( mode ){
    case tString     : arg.reset( new StrMatrix() ); break;
    case tMpMatrix   : arg.reset( new MpMatrix()  ); break;
    case tCMpMatrix  : arg.reset( new CMpMatrix() ); break;
    case tRatMatrix  : arg.reset( new RatMatrix() ); break;
    case tDMatrix    : arg.reset( new DMatrix()   ); break;
    case tBMatrix    : arg.reset( new BMatrix()   ); break;
    case tVoid       : throw InvalidOperation("Void type not supported yet due to major refactoring" );      break;
    case tUnknown    : throw InvalidOperation("Error assigning to Unknown type");   break;    
    default          : throw InvalidOperation("Error assigning to INVALID TYPE" );  break;
  }
}


//TODO tComplex, tRational, ... , MpMatrix, iterate and show verbose...
Matrix* getVerbose( const MatrixPtr& arg ){
  string verbString = "";
  
  if( currentMode == tMpMatrix ){
    verbString = (*(MpMatrix*)arg.get())(1).getVerbose();
  }
  
  return new StrMatrix( verbString );
  
}

======================================================*/



